/*
** AnimatedGraph2D.c
**
** Animated 2D function graph
** 
** 07/02/06 v1.0 Lucio Di Jasio
** 11/21/07 v2.0 LDJ PIC32 porting and animation
** 
** HW req.: Explorer16 board + AV1632 with NTSC/PAL video
** SW req.: explore.c graphic.c 
**
** This is a modified version of the Graph2D project presented
** in the book (Chapter 12) with added real time animation.
** It demonstrates 2D (float) function graphing on wire frame.
** The smoothness of the animation (thanks to double buffering)
** and 400 points redraw is a testimony of the PIC32 speed.
** Usage:
**  After start up, press a button to draw the picture
**  then use the four buttons to modify the image
**     - 1 and 2 (from left) control lateral movement
**     - 3 and 4 (from left) control amplitude
*/

// configuration bit settings
#pragma config POSCMOD=XT, FNOSC=PRIPLL
#pragma config FPLLIDIV=DIV_2, FPLLMUL=MUL_18, FPLLODIV=DIV_1
#pragma config FWDTEN=OFF, CP=OFF, BWP=OFF

#include <p32xxxx.h>
#include <explore.h>
#include <graphic.h>
#include <math.h>

#define X0      10              // graph offset
#define Y0      10
#define NODES   20              // define grid 
#define SIDE    10              
#define STEP    1               // movement increment

typedef struct {
        int x;
        int y;
    } point;

point edge[NODES], prev;

main( void)
{
    int i, j, x, y, z;
    float xf, yf, zf, sf;
    int px, py;
    int xoff, scale;

    // initializations
    initEX16();
    clearScreen();
    initVideo();    

    xoff = 100;

    if ( getKEY() == 8)     // first button show entire graph
        scale = 75;
    else
        scale = 0;

    while (1)
    {
        // clear hidden screen 
        clearHScreen();

        // draw the x, y and z axes crossing in (X0,Y0)
        line( X0, 10, X0, 10);              // z axis
        line( X0-5, Y0, HRES-10, Y0);       // x axis 
        line( X0-2, Y0-2, X0+120, Y0+120);  // y axis
    
        // init the array of previous egde points
        for( j = 0; j<NODES; j++)
        {
            edge[j].x = X0+ j*SIDE/2;
            edge[j].y = Y0+ j*SIDE/2;
        }
    
        // plot the graph of the function for 
        for( i=0; i<NODES; i++)
        {
            // transform the x coordinate range to 0..200 offset 100
            x = i * SIDE;
            xf = (6 * M_PI / 200) * (float)(x-xoff);
            prev.y = Y0;
            prev.x = X0 + x;
    
            for ( j=0; j<NODES; j++)
            { 
                // transform the y coordinate range to 0..200 offset 100
                y = j * SIDE;
                yf = (6 * M_PI / 200) * (float)(y-100);
                
                // compute the function
                sf = sqrt( xf * xf + yf * yf);
                zf = 1/(1+ sf) * cos( sf );
                
                // scale the output
                z = zf * scale;
                
                // apply isometric perspective and offset
                px = X0 + x + y/2;
                py = Y0 + z + y/2; 
                
                // plot the point
                plot( px, py);
            
                // draw connecting lines to visualize the grid
                line( px, py, prev.x, prev.y);  // connect to prev point on same x
                line( px, py, edge[j].x, edge[j].y);
    
                // update the previous points
                prev.x = px;
                prev.y = py;
                edge[j].x = px;
                edge[j].y = py;
            } // for j
        } // for i
        
        // make the drawing visible
        swapV();

        // wait for a button 
        switch( readKEY())
        {
        case 8:
            xoff -= STEP;
            break;
        case 4:
            xoff += STEP;
            break;
        case 2:
            scale -= STEP;
            break;
        case 1:
            scale += STEP;
            break;
        }        
    } // main loop

} // main
