/*
** Mandelbrot3.c
**
** Mandelbrot Set Advanced Graphic Demo
**
** 8/16/06 rev 3.0  Lucio Di Jasio
** 8/08/07 rev 3.1  LDJ PIC32 porting 
**
** HW req.: Explorer16 board + Av1632 with NTSC/PAL video
**               and SD/MMC card interface
** SW req.: Explore.c sdmmc.c fileio.c graphic.c 
** Ref:     www.explorepic32.com
**
** This demo is an enhanced version of the Mandelbrot demo 
** presented in chapter 13 with added options to save the
** current display content in a screenshot (SVR) file
** self timing for (float) performance comparisons and 
** offering additional navigation features:
**  - buttons 1, 2, 3 and 4 (from left)
**      zoom in corresponding quadrant
**  - buttons 1 + 2 
**      zoom in center
**  - buttons 1 + 4
**      save screenshot to .SVR file
**  - button 1 (press and keep pressing for >2s)
**      re-start from cardiod 
*/

// configuration bit settings
#pragma config POSCMOD=XT, FNOSC=PRIPLL
#pragma config FPLLIDIV=DIV_2, FPLLMUL=MUL_18, FPLLODIV=DIV_1
#pragma config FWDTEN=OFF, CP=OFF, BWP=ON

#include <p32xxxx.h>
#include <plib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <explore.h>
#include <graphic.h>
#include <sdmmc.h>
#include <fileio.h>

#define SIZE  VRES
#define MAXIT 64

void mandelbrot( float xx0, float yy0, float w)
{
    float x, y, d, x0, y0, x2, y2;
    int i, j, k;

    // calculate increments
    d = w/SIZE;
    
    // repeat on each screen pixel
    y0 = yy0;
    for (i=0; i<SIZE; i++)
    {
        x0 = xx0;
        for (j=0; j<SIZE; j++)
        {        
            // initialization
              x = x0;
              y = y0;         
              k = 0;
             
             // core iteration
              do {
                x2 = x*x;
                y2 = y*y;
                y = 2*x*y + y0;
                x = x2 - y2 + x0;                        
                k++;
              } while ( (x2 + y2 < 4)  && ( k < MAXIT)); 
            
            // check if the point belongs to the Mandelbrot set
            if ( k & 1) plot( j, i);
            //if ( k == MAXIT) plot( j, i);

            // compute next point x0
            x0 += d;             
         } // for j
         PORTA = i;
         // compute next y0
         y0 += d; 

        // check for a break 
        if (readKEY())
        {   
            getKEY();
            return;
        }
     } // for i
 } // mandelbrot
 

int saveScreen( void)
{
    MFILE *fp;
    char s[]= "MAND0000.SVR";
    int r, w;

    haltVideo();
    // mount SD card if present and available
    if ( !mount())
        return -1;
    
    // generate a random name for the new video image
    // retry if the file already exist
    do {
        r = rand()%9999;
        sprintf( &s[4], "%04d", r);
        strcat( s, ".SVR");  
        fp = fopenM( s, "w");
    } while( fp == NULL);

    // write screen content to file
    w = fwriteM( (char *)getAVideo(), VRES*HRES/8, fp);

    // close the file and unmount SD card
    fcloseM( fp);
    unmount();
    initVideo();

    //check for writing errors
    if ( w != VRES*HRES/8)
        return -2;

    // if successful return the file name/number
    return r;
} // saveScreen


int main()
{
    double x, y, w;
    double ox, oy, ow;
    int c, oc, k, r, recalc;
    unsigned t;
    
    // initializations
	initEX16();     // initialize the Explorer16 board
    initVideo();    // start the state machines
    singleV();
    OpenTimer45( T4_ON | T4_SOURCE_INT, 0);
    
    AT(6, 10);  putsV( " Mandelbrot Set Demo");
    AT(6, 12);  putsV( "      v.3.1");
    AT(6, 16);  putsV( "(press a key to start)");

    while( !readKEY()); // randomizing with the first button press
    srand( TMR1);
    getKEY();

    // intial coordinates lower left corner of the grid 
    x = -2.0;
    y = -2.0;
    // initial grid size
    w = 4.0;
    c = 1;
    
    char s[16];
	while( 1)
	{    
        clearScreen();          // clear the screen
        AT( 25, 0); sprintf( s, "%+.4f", x); putsV( s);
        AT( 25, 2); sprintf( s, "%+.4f", y); putsV( s);
        AT( 25, 4); sprintf( s, "%+.4f", w); putsV( s);

        AT( 25, 6); sprintf( s, "x%d", c); putsV( s);

        WriteTimer45( 0);
        
        mandelbrot( x, y, w);   // draw new image
        
        t = ReadTimer45();
        AT ( 25, 8); sprintf( s, "%1.4f", t/36000000.0 );
        putsV( s);
//        CloseTimer45();
        
        // wait for a command
        do {
            recalc = TRUE;
            k = getKEY();
            // undo button, press and hold the rightmost button
            if (k == 0x81)
            {
                x = ox; y = oy;
                w = ow; c = oc;
            }          
            else    
            {   // save values for undo
                ox = x; oy = y;       
                ow = w; oc = c;
                    
                switch( k)
                {  // select function   
                case 8:            // first quadrant
                    w/= 2;
                    y += w;
                    c<<=1;
                    break;
                
                case 4:            // second quadrant
                    w/= 2;
                    y += w;
                    x += w;
                    c<<=1;
                    break;
                
                case 2:            // third quadrant
                    w/= 2;
                    x += w;
                    c<<=1;
                    break;
                
                case 1:            // fourth quadrant
                    w/= 2;
                    c<<=1;
                    break;
                        
                // first and second button                        
                case 0xc:          // zoom in center
                    x += w/2;
                    y += w/2;
                    w /=2;
                    c<<=1;
                    break;

                //first and last button                    
                case 9:            // save to flash media
                    r = saveScreen();
                    AT( 25, 8); putsV( "Saved!");
                    AT( 25, 10); sprintf( s, "%d", r); putsV( s);
                    recalc = FALSE;
                    break;
    
                // press and hold first button
                case 0x88:          // reset origin
                    x = -2.0;
                    y = -2.0;
                    w = 4.0;
                    c = 1;
                    break;
                    
                default:
                    recalc = FALSE;
                    break;  
                } // switch
            } // else
        } while( !recalc);
	} // main loop

} // main
