/*
**   WAVE Player with Graphic Equalizer
**
**	8/20/06 v2.0 Lucio Di Jasio
**  7/24/07 v2.1 LDJ PIC32 porting
** 11/22/07 v2.2 LDJ Graphic Equalizer
**
** HW req.: Explorer16 board + AV1632 with NTSC/PAL video
**              + SD/MMC interface + Audio interface
** SW req.: explore.c graphic.c sdmmc.c fileio.c fft.c
**           equalizer.c audioPWMB.c
** ref:     www.exploringpic32.com
**
** Demonstrates:
**  -  use of PWM for stereo sound reproduction
**  -  fast fileio from SD/MMC card
**  -  Wave file decoding
**  -  DSP performance (FFT)
**  -  real time graphics animated (single buffered)
**
** After start up will prompt the user to insert a card 
** then will create a list of WAVE files present (in root)
** up to 40 files, and sorts them alphabetically
** Four commands are available for playback:
**  - button 1 (leftmost): play previous in list
**  - button 2 : play next in list
**  - button 3 : play all sequentially (alphabetically)
**  - button 4 : play all in random order (shuffle)
**
** press any button during playback to interrupt.
*/

// configuration bit settings, Fcy=72MHz, Fpb=36MHz
#pragma config POSCMOD=XT, FNOSC=PRIPLL 
#pragma config FPLLIDIV=DIV_2, FPLLMUL=MUL_18, FPLLODIV=DIV_1
#pragma config FPBDIV=DIV_2, FWDTEN=OFF, CP=OFF, BWP=OFF

#include <p32xxxx.h>
#include <plib.h>

#include <stdlib.h>     // rand, NULL...
#include <string.h>     // memcpy...
#include <stdio.h>      // sprintf

#include <explore.h>
#include <SDMMC.h>
#include <fileio.h>
#include <graphic.h>
#include "fft.h"
#include "SuperAudioPWM.h"

#define LIST_MAX    40
char list[LIST_MAX][8];        // list of file names matching
char name[13];
char s[16];

int comp( const void *a, const void *b)
{ // inverse alphabetical order
    return -memcmp( a, b, 8); 
} // comp



unsigned play( int i)
{
    // extract the name from the list
    strncpy( name, list[i], 8); name[8] = '\0';
    // update display
    AT( 18, 1); putsV( "Playing:");
    AT( 18, 2); putsV( name);

    // form a complete file name
    strcat( name, ".WAV");

    // play the file, terminate on error or key pressed
    return  playWAV( name);

} // play



int main( void)
{
    short int i, n, c;
    short int Index;

    // init the FFT rotation and bit reversal arrays
    initFFT();

    // init Explorer16 board    
    initEX16();
        
    // init modules
    initAudio();
    initVideo();
    singleV();
    clearScreen();
    
    // logo
    AT(6,10); putsV( "  PIC32 Media Player");
    AT(6,12); putsV( "        v3.0");
    AT(6,14); putsV( "S1->Prev     S2->Next");
    AT(6,16); putsV( "S3->PlayAll  S4->Shuffle");
    
    // randomize
    while( !readKEY());
    srand( TMR1);
    getKEY();

    // repeatedly try to mount the storage device
    do{
        clearScreen();
        putsV( "Insert SDCard...");

        // wait for card to be inserted 
        while( !getCD());
        // and contact debounce
        Delayms( 100);
    } while ( !mount());
   
    // clear LCD display 
    putsV( "Card Detected");
    putsV( "Creating Playlist");

    // loop 
    while( 1) // main loop
    {
        // create a list of all WAV files on the root
        n = listTYPE( (char *)list, LIST_MAX, "WAV");

        // show how many songs on the first line
        sprintf( s, "Found: %d songs", n);
        putsV( s);

        if ( n < 1)
        { 
            unmount();     
            putsV( "No WAV files found!");
            putsV( "Change Card, then press any button...");
            getKEY();      // wait for a key
            clearScreen();
            continue;       // back to the top of the loop
        }
        else
            Index = n;

        // sort the list
        qsort( list, n, 8, comp);


        // wait for a command, play prev., play next, seq., shuffle
        do {

          switch ( (c = getKEY()))
          {
          case 2:
          case 1:
            // play all WAV files sequential or shuffle    
            while(( n>0))
            {
                clearScreen();

                // select next file in shuffle mode or sequential mode
                if ( c == 1)    // shuffle 
                {
                    i = rand() % n;
                    sprintf(s, "Shfl Mode: %d", i);
                    putsV( s);
                }
                else            // sequential
                {
                    i = n-1;
                    sprintf(s, "Seq Mode: %d", i);
                    putsV( s);
                }

                // ask to press a key to get started
                //putsV( "Press a key to start");
                //getKEY();
                while( readKEY());
                
                // form a complete file name
                if ( !play( i))
                    break;  

                // remove file name from the list
                if ( i < n-1)
                    memcpy( list[i], list[n-1], 8);
    
                // remove last file from the list
                n--;                   
                } // while last 
            break; // sequential or shuffle

         case 8:  // previous file single
            clearScreen();
            if ( --Index <0)    
                Index = n-1;
            AT (18, 0); sprintf(s, "Index: %d", Index);
            putsV( s);
            // play the file, terminate on error or key pressed
            play( Index);
            break; 

         case 4: // next file single
            clearScreen();
            if ( Index++ >= n)
                Index = 0;
            AT( 18,0); sprintf(s, "Index: %d", Index);
            putsV( s);

            // play the file, terminate on error or key pressed
            play( Index);
            break;

         default:
            break;
        } //switch
      } while ( n>0); // wait for command
    } // main loop
} //main 
