/*
** AudioPWMB.c
**
**  7/22/06 v1.1  Lucio Di Jasio
** 07/14/07 v2.0 LDJ porting to PIC32
** 11/22/07 v2.1 LDJ integrate WAVE decoding and background task
*/

#include <p32xxxx.h>
#include <plib.h>
#include <explore.h>
#include <stdlib.h>

#include <sdmmc.h>
#include <fileio.h>
#include <graphic.h>
#include "SuperAudioPWM.h"
#include "fft.h"

// global definitions
unsigned short Offset;              // 50% duty cycle value
char    ABuffer[ 2][ B_SIZE];       // double data buffer
volatile short CurBuf;              // index of buffer in use
volatile short AEmptyFlag;          // flag a buffer needs to be filled

// audio configuration parameters
AudioCfg ACfg;       // audio configuration parameters

// local definitions
unsigned char *BPtr;                // pointer inside active buffer
short BCount;                       // count bytes used 


void initAudio( void)
{ // configures peripherals for Audio playback 
    
    // 1. activate the PWM modules 
    // CH1 and CH2 in PWM mode, TMR2 based
    OpenOC1( OC_ON | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE,
            0, 0);
    OpenOC2( OC_ON | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE,
            0, 0);

    // 2. init the timebase
    // enable TMR2, prescale 1:1, internal clock, period
    OpenTimer2(T2_ON | T2_PS_1_1 |  T2_SOURCE_INT, 0);         
    mT2SetIntPriority( 4);      // set TMR2 interrupt priority
    
} // initAudio


void startAudio( int bitrate, int position, int count)
{ // begins the audio playback
    
    // 1. init pointers  and flags
    CurBuf = 0;                 // buffer 0 active first
    BPtr = ABuffer[ CurBuf] + position;
    AEmptyFlag = 0;

    // 2. number of samples to be played
    BCount = count; 

    // 3. set the period for the given bitrate
    PR2 = FPB / bitrate;    

    // 4. set the initial duty cycles
    Offset = 0x80;
    OC1R = OC1RS = Offset;  // left
    OC2R = OC2RS = Offset;  // right
    
    // 5. enable the interrupt state machine
    mT2ClearIntFlag();             // clear interrupt flag
    mT2IntEnable( 1);              // enable TMR2 interrupt
} // startAudio


void haltAudio( void)
{ // stops playback state machine
   mT2IntEnable( 0);
} // halt audio


void __ISR(  _TIMER_2_VECTOR, ipl4) T2Interrupt( void)
{
    // 0. allow interrupt nesting
    asm( "ei");
    
    // 1. load the new samples for the next cycle
    OC1RS = 30+(*BPtr ^ ACfg.fix); 
    if ( ACfg.stereo)
        OC2RS = 30 + (*(BPtr + ACfg.size) ^ ACfg.fix);
    else    // mono
        OC2RS = OC1RS;
        
    // 1.1 if requested fill the fft input buffer
    if ( backgroundState == S_SAMPLE_REQ)
    {
        inB[inCount++] = OC1RS;
        // if input buffer filled advance state
        if (inCount>=N_FFT)
            backgroundState++;
    }

    // 2. skip samples to reduce the bitrate
    BPtr += ACfg.skip;

    // 3. check if buffer emptied
    if ( --BCount == 0)
    {
        // 3.1 swap buffers
        CurBuf = 1- CurBuf;
        BPtr = &ABuffer[ CurBuf][ACfg.size-1];
        
        // 3.2 restart counter
        BCount = B_SIZE/ACfg.skip;

        // 3.3 flag a new buffer needs to be filled
        AEmptyFlag = 1;
    }

    // 4. clear interrupt flag and exit
    mT2ClearIntFlag();
} // T2Interrupt

/*-------------------------------------------------------------
** wave display
*/
void waveDisplay( unsigned char *b)
{  
    int i, j;
    char *p = getAVideo()+ (VRES-N_FFT-1)*32+16;
    
    for (i=0; i<N_FFT; i++)
    {
        for( j=0;  j<16; j++)
            p[j] = 0;
        p+=32;
        plot( 128+ *b++/2, N_FFT-i);
    }
} // waveDisplay

/*-------------------------------------------------------------
** background task manager
*/
volatile int backgroundState = S_SAMPLE_REQ;
volatile int inCount = 0;

void background( void)
{
    switch( backgroundState){
    default:
    case S_SAMPLE_REQ:
        // wait the interrupt routine will advance it
        break;
    case S_WINDOW_FFT:
        windowFFT(inB);
        FFT();
        backgroundState = S_SCALE;
        break;
    case S_SCALE:
        inCount=0;
        backgroundState = S_SAMPLE_REQ;
        powerScale(outB);
        equDisplay( outB);
        while ( backgroundState==S_SAMPLE_REQ);
        waveDisplay( inB);
        break;
    } // switch
} // background


/*-------------------------------------------------------------
** Wave.C
** 
** 	Wave File Player 
**	Uses 2 x 8 bit PWM channels at 44kHz
**
** revision:
** 05/02/03 v1.00  LDJ PIC18
** 07/08/06 v2.00  LDJ porting to PIC24/C30
** 07/14/07 v3.00  LDJ porting to PIC32
** 10/25/07 v3.10  LDJ added graphic support
*/

// WAVE file constants
#define  RIFF_DWORD  0x46464952UL
#define  WAVE_DWORD  0x45564157UL
#define  DATA_DWORD  0x61746164UL
#define  FMT_DWORD   0x20746d66UL     
#define  WAV_DWORD   0x00564157UL

typedef struct {
    // data chunk
    unsigned int dlength;       // actual data size
    char    data[4];            // "data"

    // format chunk
    unsigned short  bitpsample; //
    unsigned short  bpsample;   // bytes per sample (4 = 16 bit stereo)
    unsigned int    bps;        // bytes per second
    unsigned int    srate;      // sample rate in Hz
    unsigned short  channels;   // number of channels (1= mono, 2= stereo)
    unsigned short  subtype;    // always 01
    unsigned int    flength;    // size of encapsulated block (16)
    char    fmt_[4];            // "fmt_"

    char    type[4];            // file type name "WAVE"
    unsigned int    tlength;    // size of encapsulated block
    char    riff[4];            // envelope "RIFF"
} WAVE; 


      
unsigned playWAV( const char *name)
{
    int     last;
    WAVE    wav;
    MFILE    *f;
    unsigned short wi;
    int lc, r, d;
    int pos;
    char s[16];
    
    // 1. open the file           
    if ( (f = fopenM( name, "r")) == NULL)
    {   // failed to open 
        return FALSE;
    }
    
    // 2. verify it is a RIFF formatted file
    if ( ReadL( f->buffer, 0) != RIFF_DWORD)
    {
        fcloseM( f);
        return FALSE;
    }
    //putrs( "RIFF file\r");
    
    // 3. look for the WAVE chunk signature 
    if ( (d = ReadL( f->buffer, 8)) != WAVE_DWORD)
    {
        fcloseM( f);
        return FALSE;
    }
    //putrs( "WAVE type chunk found\r");
    
    // 4. look for the chunk containing the wave format data
    if ( ReadL( f->buffer, 12) != FMT_DWORD)
    {
        fcloseM( f);
        return FALSE;
    }
    //putrs( "Format chunk contains:\r");
    
    wav.channels    = ReadW( f->buffer, 22);
    wav.bitpsample  = ReadW( f->buffer, 34);
    wav.srate       = ReadL( f->buffer, 24);
    wav.bps         = ReadL( f->buffer, 28);
    wav.bpsample    = ReadW( f->buffer, 32);
    
    //printb( "Channels        = ", wav.channels); pcr();                        
    //printi( "Bits per Sample = ", wav.bitpsample); pcr();                        
    //printl( "Sample Rate     = ", wav.srate); pcr();                        
    //printl( "Bytes per Second= ", wav.bps); pcr();                        
    //printi( "Bytes per Sample= ", wav.bpsample); pcr();                        

    // 5. search for the data chunk
    wi = 20 + ReadW( f->buffer, 16);  
    while ( wi < 512)
    {
        if (ReadL( f->buffer, wi) == DATA_DWORD)
            break;
        wi += 8 + ReadW( f->buffer, wi+4);
    }
    if ( wi >= 512) // could not find a data chunk in current sector
    {
        //putrs( " DATA chunk not found in first sector!\r");
        fcloseM( f);
        return FALSE;
    }   
    //printxi( "DATA chunk found at offset: 0x", wi+8); pcr();

    // 5.1 find the data size (actual wave content)
    wav.dlength = ReadL( f->buffer, wi+4);
    
    // 6. compute the period and adjust the bit rate
    //printl( "Data size       = ", wav.dlength); pcr();
    r = wav.bps / wav.bpsample;    // r = samples per second
    ACfg.skip = wav.bpsample;           // bytes per sample x chn
    
    // limit the maximum sample rate to 48k
    while ( r > 48000)
    {
        r >>= 1;                   // divide sample rate by two
        ACfg.skip <<= 1;                // multiply skip by two
    }
    //printl( "Playing at = ", (dword)r); putrs( " samples per second\r");

    // 6.1 determine timer period required
    d = (FPB/r)-1;
    if ( d > ( 65536L))             // max period value(16-bit)
    {
        //printi("Too long period = ", rate); pcr();
        fcloseM( f);
        return FALSE;
    }
    //printi( "TMR2 period= ", d); pcr();

    // 7. init the Audio state machine
    CurBuf = 0;
    ACfg.stereo = (wav.channels == 2);
    pos   = wi+8;                   // data begin 
    ACfg.size  = 1;                 // bytes per channel
    ACfg.fix   = 0;                 // sign fix / 16 bit file
    if ( wav.bitpsample == 16)
    {
        pos++;                      // add 1 to get the msb
        ACfg.size = 2;
        ACfg.fix  = 0x80;
    }
    
    // 8. start loading the buffers
    // compute lba of first sector 
    //l = dsk->data + (dword)(f.ccls-2) * dsk->sxc;
    // determine the number of bytes to composing the wav data chunk
    lc = wav.dlength;
    
    // 9. pre-load first buffer
    r = freadM( ABuffer[0], B_SIZE, f);
    if ( r < B_SIZE)                // file short
       lc = 0;                      // entire file used up
    else 
        lc-= B_SIZE ;                  
    AEmptyFlag = TRUE;              // immediately ask for more

    // 9.1 print WAV detail
    AT( 18, 4); putsV( "Length:");
    AT( 19, 5); sprintf( s, "%d byte", wav.dlength); putsV(s);
    AT( 18, 6); putsV( "Sample Rate:");
    AT( 19, 7); sprintf( s, "%d smp/s", wav.srate); putsV(s);
    AT( 18, 8); putsV( "Resolution:");
    AT( 19, 9); sprintf( s, "%d bit", wav.bitpsample); putsV(s);
    AT( 19,10); 
    if (wav.channels=2)
        putsV( "Stereo");
    else
        putsV("Mono");
    
    // 10. configure Player state machine and start
    startAudio( wav.srate, pos, (r-pos)/ACfg.skip);
        
    //putrs( "Playing now...\r");
    //CTS_USART1 = 0;                 // hardware handshake open 

    maxScale = 0.0;

    // 10. keep feeding the buffers in the playing loop
    while (lc >=B_SIZE)
    {
        if (readKEY())              // on pressing any key
        {                           
            lc = 0;                 // signal playback completed
            break;
        }
        if ( AEmptyFlag)
        {
            r = freadM( ABuffer[1-CurBuf], B_SIZE, f);
            AEmptyFlag = FALSE;
            lc-= B_SIZE;            

         // put here extra comp. loads
            background();
        }
    } // while wav data available

    // 11. flush the buffers with the data tail
    if( lc>0) 
    {
        // load the last sector 
        r = freadM( ABuffer[1-CurBuf], lc, f);
        last = ABuffer[1-CurBuf][r-1];
        while(( r<B_SIZE) && (last>0)) 
            ABuffer[1-CurBuf][r++] = last;
        
        // wait for current buffer to be emptied
        AEmptyFlag = FALSE;
        while (!AEmptyFlag);
    }

    // 12.finish the last buffer
    AEmptyFlag = FALSE;
    while (!AEmptyFlag);
    
    // 13. stop playback 
    haltAudio();

    // 14. close the file 
    fcloseM( f);

    // 15. return with success
    return TRUE;

} // playWAV
         


