/*
**  Slideshow.c
**
**	9/14/07 v2.0 LDJ  PIC32 porting  
**
** HW req.: Explorer16 board + AV1632 with NTSC/PAL video
**              + SD/MMC interface + Audio interface
** SW req.: explore.c graphic.c sdmmc.c fileio.c 
** ref:     www.exploringpic32.com
**
** Demonstrates:
**  -  fast fileio from SD/MMC card
**
** After start up will prompt the user to insert a card 
** then will create a list of .SVR files present (in root)
** up to 40 files, and sorts them alphabetically
** Four commands are available for slideshow playback:
**  - button 1 (leftmost): play previous in list
**  - button 2 : play next in list
**  - button 3 : play all sequentially (alphabetically)
**  - button 4 : play all in random order (shuffle)
**
** press any button during playback to interrupt.
*/


// configuration bit settings
#pragma config POSCMOD=XT, FNOSC=PRIPLL
#pragma config FPLLIDIV=DIV_2, FPLLMUL=MUL_18, FPLLODIV=DIV_1
#pragma config FWDTEN=OFF, CP=OFF, BWP=ON

#include <p32xxxx.h>
#include <stdlib.h>


#include <sdmmc.h>
#include <fileio.h>
#include <graphic.h>
#include <explore.h>

#define PAUSE       4           // delay between slides
#define LIST_MAX    40          // max number of slides
char list[LIST_MAX][8];         // list of file names matching
char aIx[LIST_MAX];             // sorting array for seq and shuffle

char s[16];

int show( int i)
{
    MFILE *fp;
    char name[13];
    int r;

    // extract the name from the list
    strncpy( name, list[i], 8); name[8] = '\0';

    // form a complete file name
    strcat( name, ".SVR");
    //putsV( name);
    fp = fopenM( name, "r");
    if ( fp == NULL)
        return FALSE;    
    else
    {
        r = freadM( getAVideo(), VRES*HRES/8, fp);
        fcloseM( fp);
    }

    // annotate the index in the right bottom corner
    AT(30, 24); sprintf( s, "%02d", i); putsV(s) ;

    return  TRUE;               // continue 
} // show


int main( void)
{
    int t, m, n, c;
    int Index, i, loop;

    // init I/O
    initEX16();
    initSD();
    initVideo();
    singleV();

    while( 1) // main loop
    {
        // splash screen
        clearScreen();
        AT( 6, 10); putsV( "     Slide Show v2.0      ");
    
        AT( 6, 14); putsV( "  Insert an SD/MMC card   ");
        AT( 6, 16); putsV( "and press any key to start");
    
        // wait for a button to start and randomize
        T1CON = 0x8000;          // init TMR1
        while( !readKEY());     // wait for a button
        srand( TMR1);           // randomize
        getKEY();               // wait for button release
            
        // wait for card to be inserted 
        if ( !getCD())
           continue;    // back to the main loop 
                
        // try to mount the storage device
        if ( !mount())
            continue;   // back to the main loop 
    
        // now look for SVR files
        clearScreen();

        // create a list of all svr files on the root
        n = listTYPE( (char *)list, LIST_MAX, "SVR");
        
        // no file found or error reading the card
        if ( n == 0)
        { 
            AT( 6, 10); putsV( "No SVR files found");
            AT( 6, 14); putsV( " (press any key)");
            unmount();      // 
            getKEY();       // wait for a key to continue
            continue;        // back to the top of the loop
        }
        
        // one or more files found
        else    
            Index = 0;

        // Show how many files where found
        AT( 8,10); sprintf( s, "%d files found", n); putsV( s);
        AT( 8,14); putsV( "+prev +next +seq +shfl");

        // wait for a command, play prev., play next, seq., shuffle
        do {
          switch ( (c = getKEY()))
          {
          case 2:
          case 1:  // play all WAV files sequential or shuffle    
            // build an index array
            for( t=0; t<n; t++)
                  aIx[t] = t;
            m = n;
            loop = TRUE;
                 
            do {
                // select next file in shuffle mode or sequential mode
                if ( c == 1)    // shuffle 
                    i = rand() % m;
                else            // sequential
                    i = 0;

                // show the file
                if ( !show( aIx[i]))
                    break;  

                // pause and check for buttons
                for( t=0; t<PAUSE*10; t++)
                {
                    Delayms( 100);
                    if ( readKEY()) // check the buttons every 1/10s
                    {
                        getKEY();   // wait for button release
                        loop = FALSE;
                        break;      
                    }   
                }

                // scroll the list down
                for ( t=i; t<m-1; t++)                  
                    aIx[t] = aIx[t+1];
                
                // remove from the top of the list
                m--;                   
                // reload list when empty
                if ( m == 0)    
                {                    
                    m = n;
                    for( i=0; i<n; i++)
                          aIx[i] = i;
                }                    
            } while( loop == TRUE);
            break; // sequential or shuffle

         case 8:  // previous file single
            if ( --Index <0)    
                Index = n-1;
            // play the file, terminate on error or key pressed
            show( Index);
            break; 

         case 4: // next file single
            if ( ++Index >= n)
                Index = 0;
            // play the file, terminate on error or key pressed
            show( Index);
            break;

         default:
            break;
        } //switch
        
      } while( 1);// wait for command
    } // main loop
} //main 
